#pragma once

// C includes
#include<sys/stat.h>
// C++ includes
#include<iostream>
#include<vector>
// Dune includes
#include<dune/grid/io/file/vtk.hh>
#include<dune/pdelab/newton/newton.hh>

#include "tmixed_chns_r_cahnhilliard.hh"
#include "tmixed_chns_navierstokes.hh"
#include "tmixed_upscaledp.hh"
#include <dune/phasefield/localoperator/pf_3p_diff_estimator.hh>

#include <dune/phasefield/utility/ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES.hh>

#include "tmixed_parameters.hh"
#include "tmixed_boundary.hh"
#include <dune/phasefield/chns_base.hh>
#include "micro_space.hh"
#include <dune/phasefield/initial_fn.hh>
#include <dune/phasefield/debug/vector_debug.hh>


template<typename GV, typename GFS, typename Parameters>
class YSlice{
public:
  typedef double RF;


  typedef typename GFS::NS NS_GFS;
  typedef typename GFS::CH CH_GFS;
  typedef Dune::PDELab::Backend::Vector<CH_GFS, RF> CH_V;
  typedef Dune::PDELab::Backend::Vector<NS_GFS, RF> NS_V;

  typedef Ini_micro< ZeroInitial<GV,RF,Parameters>, //VX
                    ZeroInitial<GV,RF,Parameters>,                       //VY
                    PhiInitial_Thinstrip_Layers<GV,RF,Parameters,0>,
                    ZeroInitial<GV,RF,Parameters>,
                    PhiInitial_Thinstrip_Layers<GV,RF,Parameters,1>,
                    ZeroInitial<GV,RF,Parameters>>  Initial;

  typedef Bdry_micro<Dune::PDELab::NoDirichletConstraintsParameters, //VX=W
                    Dune::PDELab::DirichletConstraintsParameters,    //VY
                    Dune::PDELab::NoDirichletConstraintsParameters,
                    Dune::PDELab::NoDirichletConstraintsParameters,
                    Dune::PDELab::NoDirichletConstraintsParameters,
                    Dune::PDELab::NoDirichletConstraintsParameters> Bdry;

  typedef TMIXED_CHNS_NavierStokes<Parameters, CH_GFS, CH_V, NS_GFS, NS_V> NS_LOP;
  typedef TMIXED_CHNS_R_CahnHilliard<Parameters, CH_GFS, CH_V, NS_GFS, NS_V> CH_LOP;
  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;

  typedef DGF_micro<GFS, NS_V, CH_V> DgfMicro;

  typedef Dune::PDELab::GridOperator<NS_GFS,NS_GFS,NS_LOP,MBE,RF,RF,RF,typename GFS::NS_CC,typename GFS::NS_CC> NS_GO;
  typedef Dune::PDELab::GridOperator<CH_GFS,CH_GFS,CH_LOP,MBE,RF,RF,RF,typename GFS::CH_CC,typename GFS::CH_CC> CH_GO;

  typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES<CH_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> CH_LS;
  typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES<NS_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> NS_LS;

  typedef Dune::PDELab::Newton<NS_GO,NS_LS,NS_V> NS_Newton;
  typedef Dune::PDELab::Newton<CH_GO,CH_LS,CH_V> CH_Newton;

  GV& gv;
  CH_V chV, chVOld, chVUpwind;
  NS_V nsV, nsVOld, nsVUpwind;
  Initial initial;
  Bdry bdry;

  RF xpos;
  RF dxp = 1;
  RF dxpUpwind = 1;
  RF c = 1;
  RF dx;
  MBE mbe;

  RF kf = 0;

  std::unique_ptr<NS_LOP> nsLop;
  std::unique_ptr<CH_LOP> chLop;
  std::unique_ptr<DgfMicro> dgfMicro;
  std::unique_ptr<NS_GO> nsGo;
  std::unique_ptr<CH_GO> chGo;
  std::unique_ptr<NS_LS> nsLs;
  std::unique_ptr<CH_LS> chLs;
  std::unique_ptr<NS_Newton>  nsNewton;
  std::unique_ptr<CH_Newton>  chNewton;

  YSlice(GV& gv_, GFS& gfs, Parameters& param, RF xpos_, RF dx_) :
  gv(gv_), chV(gfs.ch), nsV(gfs.ns), chVOld(gfs.ch), nsVOld(gfs.ns), chVUpwind(gfs.ch), nsVUpwind(gfs.ns),
  initial(gv,param), xpos(xpos_), mbe(10), dx(dx_)
  {
    param.TempXpos = xpos;
    Dune::PDELab::interpolate(initial.iniCH , gfs.ch, chV);
    Dune::PDELab::interpolate(initial.iniNS , gfs.ns, nsV);
    gfs.updateConstraintsContainer(bdry);
    interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsV,chV);

    chVOld = chV;
    nsVOld = nsV;
    chVUpwind = chV;
    nsVUpwind = nsV;

    //Local Operator
    nsLop = std::make_unique<NS_LOP>(param, gfs.ch, chV, gfs.ns);
    chLop = std::make_unique<CH_LOP>(param, gfs.ch, chVOld, chVUpwind, gfs.ns, nsV, nsVUpwind, dxp, dxpUpwind, c, dx);

    dgfMicro = std::make_unique<DgfMicro>(gfs,nsV,chV);

// From here on each timestep (if grid changes)
    //Grid Operator
    nsGo = std::make_unique<NS_GO>(gfs.ns,gfs.nsCC,gfs.ns,gfs.nsCC,*nsLop,mbe);
    chGo = std::make_unique<CH_GO>(gfs.ch,gfs.chCC,gfs.ch,gfs.chCC,*chLop,mbe);
    // Linear solver
    chLs = std::make_unique<CH_LS>(*chGo,500,100,3,1);
    nsLs = std::make_unique<NS_LS>(*nsGo,1000,200,3,1);
    // Nonlinear solver

    nsNewton = std::make_unique<NS_Newton>(*nsGo,nsV,*nsLs);
    nsNewton->setParameters(param.nsNewtonTree);

    chNewton = std::make_unique<CH_Newton>(*chGo,chV,*chLs);
    chNewton->setParameters(param.chNewtonTree);

    printVector(chV,10,"chV");
    printVector(nsV,10,"nsV");
  }

  void applyNs()
  {
    int verb = 2;
    if (verb > 1)   std::cout << "= Begin newtonapply NS:" << std::endl;
    nsNewton->apply();
    nsVOld = nsV;
  }

  void applyCh()
  {
    int verb=2;
    if (verb > 1)   std::cout << "= Begin newtonapply CH:" << " with dxp= " << dxp << " dxpUpwind= " << dxpUpwind << " c= " << c << std::endl;
    chNewton->apply();
    chVOld = chV;
  }

  void updateKf()
  {
    auto productFunction = Dune::PDELab::makeGridFunctionFromCallable (gv, [&](const auto& element, const auto& xlocal){
        Dune::FieldVector<double,1> w;
        dgfMicro->vxDgf.evaluate(element, xlocal, w);
        Dune::FieldVector<double,1> phi1;
        dgfMicro->phiDgf.evaluate(element, xlocal, phi1);
        Dune::FieldVector<double,1> phi2;
        dgfMicro->phi2Dgf.evaluate(element, xlocal, phi2);
        return Dune::FieldVector<double,1>((phi1+phi2)*w);
    });
    Dune::FieldVector<double,1> integral;
    Dune::PDELab::integrateGridFunction(productFunction,integral,2);
    kf = integral[0];
  }
};

template<typename RF, typename YSimS>
class P_ParamAdaptor
{
public:
  const int xs;
  YSimS& ySimS;
  P_ParamAdaptor(YSimS& ySimS_) : ySimS(ySimS_), xs(ySimS_.size())
  {
  }

  RF evalKf(const auto& xg)
  {
    int i = std::max(0, std::min(xs-1, (int)(xg[0]*xs + 1e-6) ));
    return ySimS[i].kf;
  }
};

//_________________________________________________________________________________________________________________________

template< typename XGV, typename PGrad_DGF, typename SimS>
void evaluateDxP(XGV& xgv, PGrad_DGF& pGradDgf, SimS& simS)
{
  Dune::FieldVector<double,1> pgrad(0);
  Dune::FieldVector<double,1> pgradUpwind(0);
  int xs = simS.size();
  auto it = elements(xgv).begin();
  for(int i=0; i<xs; i++)
  {
    pGradDgf.evaluate(*it,it->geometry().local(it->geometry().center()), pgrad );
    simS[i].dxp = pgrad[0];
    simS[i].dxpUpwind = pgradUpwind[0];
    it++;
    pgradUpwind = pgrad;
  }
  //Periodic:
  simS[0].dxpUpwind = simS[xs-1].dxp;
}

//_________________________________________________________________________________________________________________________

template<typename GV, typename ES, typename TGV, typename T_FEM, typename V_FEM, typename P_FEM, typename PHI_FEM>
void driver_mixed(GV& xgv, const P_FEM pFem, int xsize,
                  GV& ygv, ES& y_es, TGV& tgv, const T_FEM& tFem, const V_FEM& vFem, const PHI_FEM& phiFem,
                  Dune::ParameterTree& ptree, Dune::MPIHelper& helper)
{

  const int dim = GV::dimension;
  typedef double RF;


  typedef Params_tmixed< RF,
                        //TripleWell<RF,DoubleWell_limited< RF, DoubleWell_poly<RF>, Limiter_FakeDiverging<RF> > >,
                        TripleWell<RF,DoubleWell_limited< RF, DoubleWell_8th_order_poly<RF>, Limiter_FakeDiverging<RF> > >,
                        Mobility_Const<RF>,
                        Reaction_Thinstrip<RF>,
                        VelDissipation_Quadratic<RF>
                        //VelDissipation_Quadratic_Shifted<RF>
                        > Parameters;
  Parameters param(ptree);
  TimestepManager timestepManager(ptree.sub("Time"),param.time,2);


//_________________________________________________________________________________________________________________________
// Make micro problems


  typedef GFS_micro<RF, ES, V_FEM, PHI_FEM> GFS;
  GFS gfs(y_es,vFem,phiFem);

  typedef YSlice<GV, GFS, Parameters> YSim;
  typedef std::vector<YSim> YSimS;
  YSimS ySimS;
  ySimS.reserve(xsize);
  for(int i=0; i<xsize; i++)
  {
    ySimS.emplace_back(ygv, gfs, param, ((RF)i)/((RF)xsize), ((RF)1)/((RF)xsize) );
  }


  //_________________________________________________________________________________________________________________________
  // Make macro problems
  using XES = Dune::PDELab::NonOverlappingEntitySet<GV>;
  XES x_es(xgv);

  typedef Dune::PDELab::ISTL::VectorBackend<> VectorBackend;
  typedef Dune::PDELab::ConformingDirichletConstraints ConstraintsAssembler;
  ConstraintsAssembler conP();
  typedef Dune::PDELab::GridFunctionSpace<XES, P_FEM, ConstraintsAssembler, VectorBackend> P_GFS;
  P_GFS pGfs(x_es,pFem);

  auto blambda = [](const auto& x){return true;};
  auto b = Dune::PDELab::
    makeBoundaryConditionFromCallable(xgv,blambda);
  //typedef Dune::PDELab::DirichletConstraintsParameters BdryP;
  //BdryP pBdry();
  typedef typename P_GFS::template ConstraintsContainer<RF>::Type P_CC;
  P_CC pCC;
  //pCC.clear();
  Dune::PDELab::constraints(b,pGfs,pCC);

  typedef Dune::PDELab::Backend::Vector<P_GFS, RF> P_V;
  P_V pV(pGfs);

  auto p0lambda = [&](const auto& xg)
    {return (1-xg[0])*(1-xg[0]);};
  auto p0 = Dune::PDELab::
      makeGridFunctionFromCallable(xgv,p0lambda);
  Dune::PDELab::interpolate(p0,pGfs,pV);




  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  MBE mbe(10);
  typedef P_ParamAdaptor<RF, YSimS> P_PAdaptor;
  P_PAdaptor pPAdaptor(ySimS);
  typedef TMIXED_UpscaledP<P_PAdaptor,P_FEM,RF> P_LOP;
  P_LOP pLop(pPAdaptor);

  typedef Dune::PDELab::GridOperator<P_GFS, P_GFS, P_LOP, MBE,
                                     RF, RF, RF, P_CC, P_CC> P_GO;
  P_GO pGo(pGfs, pCC, pGfs, pCC, pLop, mbe);
  typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES<P_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> P_LS;
  P_LS pLs(pGo,500,100,3,1);
  typedef Dune::PDELab::Newton<P_GO,P_LS,P_V> P_Newton;
  P_Newton pNewton(pGo,pV,pLs);
  pNewton.setReassembleThreshold(0.0); // always reassemble J
  pNewton.setVerbosityLevel(3);        // be verbose
  pNewton.setReduction(1e-10);         // total reduction
  pNewton.setMinLinearReduction(1e-4); // min. red. in lin. solve
  pNewton.setMaxIterations(25);        // limit number of its
  pNewton.setLineSearchMaxIterations(10); // limit line search
  pNewton.setAbsoluteLimit(1e-10);

  //_________________________________________________________________________________________________________________________
  // prepare VTK writer
  std::string filename=ptree.get("output.filename","output");
  std::string x_filename = "xoutput";
  struct stat st;
  if( stat( filename.c_str(), &st ) != 0 )
  {
    int stat = 0;
    stat = mkdir(filename.c_str(),S_IRWXU|S_IRWXG|S_IRWXO);
    if( stat != 0 && stat != -1)
      std::cout << "Error: Cannot create directory "  << filename << std::endl;
  }
  if( stat( x_filename.c_str(), &st ) != 0 )
  {
    int stat = 0;
    stat = mkdir(x_filename.c_str(),S_IRWXU|S_IRWXG|S_IRWXO);
    if( stat != 0 && stat != -1)
      std::cout << "Error: Cannot create directory "  << x_filename << std::endl;
  }
  auto stationaryvtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<TGV> >(tgv,0);
  Dune::VTKSequenceWriter<TGV> vtkwriter(stationaryvtkwriter,filename.c_str(),filename.c_str(),"");

  DGF_macro<RF, GV, TGV, YSimS> dgfMacro(ygv, tgv,ySimS);
  dgfMacro.addToVTKWriter(vtkwriter);
  vtkwriter.write(param.time.t,Dune::VTK::base64);

  auto x_stationaryvtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<GV> >(xgv,0);
  Dune::VTKSequenceWriter<GV> x_vtkwriter(x_stationaryvtkwriter,x_filename.c_str(),x_filename.c_str(),"");

  typedef Dune::PDELab::DiscreteGridFunction<P_GFS,P_V> P_DGF;
  P_DGF pDgf(pGfs,pV);
  typedef Dune::PDELab::DiscreteGridFunctionGradient<P_GFS,P_V> PGrad_DGF;
  PGrad_DGF pGradDgf(pGfs,pV);

  typedef Dune::PDELab::VTKGridFunctionAdapter<P_DGF> P_VTKF;
  x_vtkwriter.addVertexData(std::shared_ptr<P_VTKF>(
                             new P_VTKF(pDgf,"p")));
  x_vtkwriter.write(param.time.t,Dune::VTK::base64);

//_________________________________________________________________________________________________________________________
  while (!timestepManager.isFinished())
  {
  for(int i=1; i<xsize; i++)
    {
      ySimS[i].chVUpwind = ySimS[i-1].chV;
    }
    ySimS[0].chVUpwind = ySimS[xsize-1].chV;  //Periodic

    for(int i=0; i<xsize; i++)
    {
      ySimS[i].applyNs();
      ySimS[i].updateKf();
    }

    for(int i=1; i<xsize; i++)
    {
      ySimS[i].nsVUpwind = ySimS[i-1].nsV;
    }
    ySimS[0].nsVUpwind = ySimS[xsize-1].nsV;  //Periodic

    //For Debug
    //timestepManager.applyTimestep(vtkwriter);
    //x_vtkwriter.write(param.time.t);


    std::cout << "Begin newtonapply Pressure:" << std::endl;
    pNewton.apply();
    evaluateDxP(xgv, pGradDgf,ySimS);
    for(int i=0; i<xsize; i++)
    {
      ySimS[i].applyCh();
    }

    timestepManager.applyTimestep(vtkwriter);
    x_vtkwriter.write(param.time.t);
  }

  //_________________________________________________________________________________________________________________________
  //Constraints

  /*typedef DomainSize<RF> DS;
  DS dS(0,param.YScaling,0,2,1e-8);
  typedef Bdry_ffs_chns_r_compositeV_2d_Explicit_Domain<
              BottomBoundary<DS>,            // v_x
              AllBoundary<DS>,         // v_y
              LeftRightBoundary<DS>,                                      // p
              LeftBoundary<DS>,                                    // phi1
              NoBoundary<DS>,       // mu1
              LeftBoundary<DS>,                                    // phi2
              NoBoundary<DS>,       // mu2
              LeftBoundary<DS>>                                      // u
              Bdry;
  Bdry bdry(dS);
  gfs.updateConstraintsContainer(bdry);*/

  //_________________________________________________________________________________________________________________________
  //Coefficient vectors

  /*using CH_V = Dune::PDELab::Backend::Vector<CH_GFS, RF>;
  using NS_V = Dune::PDELab::Backend::Vector<NS_GFS, RF>;
  using CH_VS = std::vector<CH_V*>;
  using NS_VS = std::vector<NS_V*>;
  CH_VS chVS(xsize,NULL);
  NS_VS nsVS(xsize,NULL);
  for(int i=0; i<xsize; i++)
  {
    chVS[i] = new CH_V(gfs.ch);
    nsVS[i] = new NS_V(gfs.ns);
  }*/

  //CH_V chV(gfs.ch);
  //NS_V nsV(gfs.ns);

  //_________________________________________________________________________________________________________________________
  //Initial Values

/*
    typedef Ini_ffs_chns_r< VelocityInitial_InOutFlow<GV,RF,Parameters,dim>,
                      PInitial_InOutFlow<GV,RF,Parameters>,
                      PhiInitial_Thinstrip_Layers<GV,RF,Parameters,0>,
                      ZeroInitial<GV,RF,Parameters>,
                      PhiInitial_Thinstrip_Layers<GV,RF,Parameters,1>,
                      ZeroInitial<GV,RF,Parameters>,
                      UInitial_Inflow<GV,RF,Parameters> >  Initial;
    Initial initial(ygv,param);
    for(int i=0; i<xsize; i++)
    {
      Dune::PDELab::interpolate(initial.iniCH , gfs.ch, *chVS[i]);
      Dune::PDELab::interpolate(initial.iniNS , gfs.ns, *nsVS[i]);
    }
*/
  //_________________________________________________________________________________________________________________________
// Refine Intial Values
/*    for(int i=0; i<ptree.get("Refinement.maxRefineLevel",2); i++)
    {
      // mark elements for refinement
      typedef RefinementIndicator<GV,RF,PF_3P_DIFF_ESTIMATOR<CH_GFS>,CH_GFS> RIND0;
      RIND0 rInd0(gv,gfs.ch,ptree.sub("Refinement"));
      rInd0.calculateIndicator(gv,chV,false);
      rInd0.markGrid(grid,2);
      //rInd0.refineGrid(grid,chV,nsV,gfs.ch,gfs.ns,1);
      rInd0.myRefineGrid(grid,gv,chV,nsV,gfs.ch,gfs.ns,1);
      Dune::PDELab::interpolate(initial.iniCH , gfs.ch, chV);
      Dune::PDELab::interpolate(initial.iniNS , gfs.ns, nsV);
    }

    gfs.updateConstraintsContainer(bdry);
    interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsV,chV);
*/

/*    CH_VS chVSOld(xsize,NULL);
    NS_VS nsVSOld(xsize,NULL);
    for(int i=0; i<xsize; i++)
    {
      chVSOld[i] = new CH_V(gfs.ch);
      *chVSOld[i] = *chVS[i];
      nsVSOld[i] = new NS_V(gfs.ns);
      *nsVSOld[i] = *nsVS[i];
    }
*/
  //_________________________________________________________________________________________________________________________
  // LocalOperators

/*  typedef THINSTRIP_FULL_CHNS_NavierStokes<Parameters, CH_GFS, CH_V, NS_GFS, NS_V> NS_LOP;
  //typedef NS_LOP_S std::vector<NS_LOP*>;
  NS_LOP nsLop(param, gfs.ch, chV, chVOld, gfs.ns, nsVOld);
  //NS_LOP_S nsLopS(xsize,NULL);

  typedef THINSTRIP_FULL_CHNS_R_CahnHilliard<Parameters, CH_GFS, CH_V, NS_GFS, NS_V> CH_LOP;
  //typedef ch_LOP_S std::vector<ch_LOP*>;
  CH_LOP chLop(param, gfs.ch, chVOld, gfs.ns, nsV);
  //CH_LOP_S chLopS(xsize,NULL);

  for(int i=0; i<xsize; i++)
  {
    chVSOld[i] = new CH_V(gfs.ch);
    *chVSOld[i] = *chVS[i];
    nsVSOld[i] = new NS_V(gfs.ns);
    *nsVSOld[i] = *nsVS[i];
  }

  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  MBE mbe(10);// Maximal number of nonzeroes per row can be cross-checked by patternStatistics().

*/

/*
  //_________________________________________________________________________________________________________________________
  // Construct discrete Grid Functions
  typedef DGF_ffs_chns_r<GFS, NS_V, CH_V> DGF;
  DGF dgf(gfs, nsV, chV);

  //_________________________________________________________________________________________________________________________
  // prepare VTK writer
  std::string filename=ptree.get("output.filename","output");
  struct stat st;
  if( stat( filename.c_str(), &st ) != 0 )
  {
    int stat = 0;
    stat = mkdir(filename.c_str(),S_IRWXU|S_IRWXG|S_IRWXO);
    if( stat != 0 && stat != -1)
      std::cout << "Error: Cannot create directory "  << filename << std::endl;
  }
  auto stationaryvtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<GV> >(gv,0);
  Dune::VTKSequenceWriter<GV> vtkwriter(stationaryvtkwriter,filename.c_str(),filename.c_str(),"");
  dgf.addToVTKWriter(vtkwriter);

  vtkwriter.write(param.time.t,Dune::VTK::ascii);

//_________________________________________________________________________________________________________________________
  printVector(chV,10,"chV");
  printVector(nsV,10,"nsV");
*/

  //____________________________________________________________________________________________________________________
  //int i_it=0;
//  chVOld = chV;
//  nsVOld = nsV;
//  while (!timestepManager.isFinished())
//    {

/*      // Grid Operators
      typedef Dune::PDELab::GridOperator<NS_GFS,NS_GFS,NS_LOP,MBE,RF,RF,RF,typename GFS::NS_CC,typename GFS::NS_CC> NS_GO;
      NS_GO nsGo(gfs.ns,gfs.nsCC,gfs.ns,gfs.nsCC,nsLop,mbe);

      typedef Dune::PDELab::GridOperator<CH_GFS,CH_GFS,CH_LOP,MBE,RF,RF,RF,typename GFS::CH_CC,typename GFS::CH_CC> CH_GO;
      CH_GO chGo(gfs.ch,gfs.chCC,gfs.ch,gfs.chCC,chLop,mbe);

      // Linear solver
      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <CH_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> CH_LS; CH_LS ls_ch(chGo,500,100,3,1);

      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <NS_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> NS_LS; NS_LS ls_ns(nsGo,1000,200,3,1);

      // Nonlinear solver
      typedef Dune::PDELab::Newton<NS_GO,NS_LS,NS_V> PDESolver1;
      PDESolver1 ns_newton(nsGo,nsV,ls_ns);
      ns_newton.setParameters(ptree.sub("NS_Newton"));

      typedef Dune::PDELab::Newton<CH_GO,CH_LS,CH_V> PDESolver2;
      PDESolver2 ch_newton(chGo,chV,ls_ch);
      ch_newton.setParameters(ptree.sub("Phasefield_Newton"));
*/
      //_________________________________________________________________________________________________________________________
      // do time step

/*
      interpolateToBdry(gfs,initial.iniNS,initial.iniCH, nsV, chV);

      int result = timestepManager.doTimestep(ns_newton,ch_newton,vtkwriter);
      //int result = timestepManager.doOnlyNSTimestep(ns_newton,vtkwriter);
      if(result <0)
      {
        return;
      }
      else if(result == 0)
      {
        nsV = nsVOld;
        chV = chVOld;
        continue;
      }

      chVOld = chV;
      nsVOld = nsV;
      child(initial.iniCH,Dune::Indices::_0).ApplyReactionDt(param);

      //_________________________________________________________________________________________________________________________

      // mark elements for refinement
      typedef RefinementIndicator<GV,RF,PF_3P_DIFF_ESTIMATOR<CH_GFS>,CH_GFS> RIND;
      RIND rInd(gv,gfs.ch,ptree.sub("Refinement"));
      rInd.calculateIndicator(gv,chV,false);
      rInd.markGrid(grid,2);
      //rInd.refineGrid(grid,chV,chVOld,nsV,nsVOld,gfs.ch,gfs.ns,2);

      rInd.myRefineGrid(grid,gv,chV,chVOld,nsV,nsVOld,gfs.ch,gfs.ns,1);

      // recompute constraints
      gfs.updateConstraintsContainer(bdry);
      interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsV,chV);
      interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsVOld,chVOld);


      //vtkwriter.write(timestepManager.t+2e-6,Dune::VTK::ascii);

*/

//    }
}
