#pragma once

#include<dune/pdelab/constraints/common/constraints.hh>
#include<dune/pdelab/constraints/conforming.hh>

#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include<dune/pdelab/gridfunctionspace/subspace.hh>
#include<dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh>
#include<dune/phasefield/utility/pf_3p_color_dgf.hh>

#include<dune/pdelab/function/callableadapter.hh>
#include <dune/pdelab/common/functionutilities.hh>


//_________________________________________________________________________________________________________________________


template<typename RF, typename ES, typename V_FEM, typename PHI_FEM>
class GFS_micro
{
public:
  typedef Dune::PDELab::ConformingDirichletConstraints ConstraintsAssembler;
  typedef Dune::PDELab::ISTL::VectorBackend<> VectorBackend;
  typedef Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::fixed> VB_Block;
  typedef Dune::PDELab::LexicographicOrderingTag OrderingTagL;
  typedef Dune::PDELab::EntityBlockedOrderingTag OrderingTagB;

  ConstraintsAssembler conphi, conmu;

  typedef Dune::PDELab::GridFunctionSpace<ES, V_FEM, ConstraintsAssembler, VectorBackend> VX_GFS;
  VX_GFS vxGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, V_FEM, ConstraintsAssembler, VectorBackend> VY_GFS;
  VY_GFS vyGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, PHI_FEM, ConstraintsAssembler, VectorBackend> PHI_GFS;
  PHI_GFS phiGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, PHI_FEM, ConstraintsAssembler, VectorBackend> MU_GFS;
  MU_GFS muGfs;

  typedef Dune::PDELab::CompositeGridFunctionSpace<VectorBackend, OrderingTagL, VX_GFS> NS;
  NS ns;

  typedef Dune::PDELab::CompositeGridFunctionSpace<VB_Block, OrderingTagB, PHI_GFS, MU_GFS, PHI_GFS, MU_GFS, VY_GFS> CH;
  CH ch;

  typedef typename NS::template ConstraintsContainer<RF>::Type NS_CC;
  NS_CC nsCC;

  typedef typename CH::template ConstraintsContainer<RF>::Type CH_CC;
  CH_CC chCC;

  GFS_micro(const ES& es, const V_FEM& vFem, const PHI_FEM& phiFem) :
      conphi(),conmu(),
      vxGfs(es,vFem),
      vyGfs(es,vFem),
      phiGfs(es,phiFem, conphi),
      muGfs(es,phiFem, conmu),
      ns(vxGfs),
      ch(phiGfs, muGfs, phiGfs, muGfs, vyGfs),
      nsCC(),
      chCC()
  {
    vxGfs.name("Xvelocity");
    vyGfs.name("Yvelocity");
    phiGfs.name("phasefield");
    muGfs.name("potential");
  }

  template<typename Bdry>
  void updateConstraintsContainer(const Bdry& bdry)
  {
    Dune::PDELab::constraints(bdry.bdryNS,ns,nsCC, 0);
    Dune::PDELab::constraints(bdry.bdryCH,ch,chCC, 0);
  }
};

//_________________________________________________________________________________________________________________________

template<typename GFS, typename NS_V, typename CH_V>
class DGF_micro
{
public:

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::NS,Dune::TypeTree::TreePath<0> > VXSubGFS;
  VXSubGFS vxSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<VXSubGFS,NS_V> VXDGF;
  VXDGF vxDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<0> > PhiSubGFS;
  PhiSubGFS phiSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<PhiSubGFS,CH_V> PhiDGF;
  PhiDGF phiDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<1> > MuSubGFS;
  MuSubGFS muSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<MuSubGFS,CH_V> MuDGF;
  MuDGF muDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<2> > Phi2SubGFS;
  Phi2SubGFS phi2SubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<Phi2SubGFS,CH_V> Phi2DGF;
  Phi2DGF phi2Dgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<3> > Mu2SubGFS;
  Mu2SubGFS mu2SubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<Mu2SubGFS,CH_V> Mu2DGF;
  Mu2DGF mu2Dgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<4> > VYSubGFS;
  VYSubGFS vySubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<VYSubGFS,CH_V> VYDGF;
  VYDGF vyDgf;

  //typedef PF_3P_Color_DGF<typename GFS::CH,CH_V,3> ColorDGF;
  //ColorDGF colorDgf;


  DGF_micro(const GFS& gfs, const NS_V& nsV, const CH_V& chV) :
      vxSubGfs(gfs.ns),
      vxDgf(vxSubGfs,nsV),

      phiSubGfs(gfs.ch),
      phiDgf(phiSubGfs,chV),
      muSubGfs(gfs.ch),
      muDgf(muSubGfs,chV),
      phi2SubGfs(gfs.ch),
      phi2Dgf(phi2SubGfs,chV),
      mu2SubGfs(gfs.ch),
      mu2Dgf(mu2SubGfs,chV),

      vySubGfs(gfs.ch),
      vyDgf(vySubGfs,chV)

      //colorDgf(gfs.ch,chV)
  {
  }

};



template<typename RF, typename GV, typename YSimS>
class DGF_evaluator
{
public:
  const GV& ygv;
  const int xs;
  YSimS& ySimS;
  const int ys;
  typedef Dune::FieldVector<RF,1> ScalarField;
  typedef Dune::FieldVector<RF,1> LocalC;
  typedef Dune::FieldVector<RF,2> GlobalC;

  struct Pos{
    int i;
    LocalC localCoord;
    typename GV::Traits::template Codim<0> :: Iterator it;
  };

  Pos FindPosition(const auto& xyglobal)
  {
    int i = std::max(0, std::min(xs-1, (int)(xs*xyglobal[0] + 0.5) ));
    int j = std::max(0, std::min(ys-1, (int)(ys*xyglobal[1]) ));
    //std::cout << "i" << i <<"j" << j << "xs" << xs << "ys" << ys << "bouns" << ySimS.size() << std::endl;
    auto it = elements(ygv).begin();
    for(int k=0; k<j; k++){ ++it; }
    LocalC yglob(xyglobal[1]);
    LocalC localCoord =  it->geometry().local(yglob);
    return Pos{i,localCoord,it};
  }

  /*ScalarField EvalPhi(const GlobalC& xyglobal)
  {
    Pos p = FindPosition(xyglobal);
    ScalarField result(0);
    ySimS[p.i].dgfMicro->phiDgf.evaluate(*(p.it),p.localCoord,result);
    return result;
  }*/

  DGF_evaluator(const GV& ygv_, YSimS& ySimS_)
  : ygv(ygv_), ySimS(ySimS_), xs(ySimS_.size()), ys(ygv.size(0))
  {
  }
};


template<typename GV, typename RF, typename Evl, int Variable>
class MacroToMicroGridFun
  : public Dune::PDELab::GridFunctionBase<Dune::PDELab::GridFunctionTraits<
        GV, RF, 1, Dune::FieldVector<RF,1>  >,
      MacroToMicroGridFun<GV,RF,Evl,Variable>
      >
{
public:
  Evl& evaluator;
  GV gv;

  typedef Dune::PDELab::
       GridFunctionTraits<GV,RF,1,Dune::FieldVector<RF,1> > Traits;
  MacroToMicroGridFun(GV gv_, Evl& evl_): gv(gv_), evaluator(evl_)
  {
  }

  inline void evaluate (const typename Traits::ElementType& e,
                        const typename Traits::DomainType& x,
                        typename Traits::RangeType& y) const
  {
    auto p = evaluator.FindPosition(e.geometry().global(x));
    switch(Variable)
    {
      case 0: evaluator.ySimS[p.i].dgfMicro->phiDgf.evaluate(*(p.it),p.localCoord,y);break;
      case 1: evaluator.ySimS[p.i].dgfMicro->muDgf.evaluate(*(p.it),p.localCoord,y);break;
      case 2: evaluator.ySimS[p.i].dgfMicro->phi2Dgf.evaluate(*(p.it),p.localCoord,y);break;
      case 3: evaluator.ySimS[p.i].dgfMicro->mu2Dgf.evaluate(*(p.it),p.localCoord,y);break;
      case 4: evaluator.ySimS[p.i].dgfMicro->vyDgf.evaluate(*(p.it),p.localCoord,y);break;
      case 5: evaluator.ySimS[p.i].dgfMicro->vxDgf.evaluate(*(p.it),p.localCoord,y);break;
    }
  }

  inline const GV& getGridView () const {return gv;}
};


template<typename RF, typename GV, typename TGV, typename YSimS>
class DGF_macro
{
public:
  typedef DGF_evaluator<RF,GV,YSimS> Evl;
  Evl evaluator;
  MacroToMicroGridFun<TGV,RF,Evl,0> phiGridFun;
  MacroToMicroGridFun<TGV,RF,Evl,1> muGridFun;
  MacroToMicroGridFun<TGV,RF,Evl,2> phi2GridFun;
  MacroToMicroGridFun<TGV,RF,Evl,3> mu2GridFun;
  MacroToMicroGridFun<TGV,RF,Evl,4> vyGridFun;
  MacroToMicroGridFun<TGV,RF,Evl,5> vxGridFun;

  DGF_macro(const GV& ygv_, const TGV& tgv_,  YSimS& ySimS_)
  : evaluator(ygv_,ySimS_),
  phiGridFun(tgv_,evaluator),
  muGridFun(tgv_,evaluator),
  phi2GridFun(tgv_,evaluator),
  mu2GridFun(tgv_,evaluator),
  vyGridFun(tgv_,evaluator),
  vxGridFun(tgv_,evaluator)
  {
  }

  template<typename VTKWRITER>
  void addToVTKWriter(VTKWRITER& vtkwriter)
  {
    vtkwriter.addVertexData(std::make_shared<
      Dune::PDELab::VTKGridFunctionAdapter<MacroToMicroGridFun<TGV,RF,Evl,0>> >(phiGridFun,"phi"));
    vtkwriter.addVertexData(std::make_shared<
      Dune::PDELab::VTKGridFunctionAdapter<MacroToMicroGridFun<TGV,RF,Evl,1>> >(muGridFun,"mu"));
    vtkwriter.addVertexData(std::make_shared<
      Dune::PDELab::VTKGridFunctionAdapter<MacroToMicroGridFun<TGV,RF,Evl,2>> >(phi2GridFun,"phi2"));
    vtkwriter.addVertexData(std::make_shared<
      Dune::PDELab::VTKGridFunctionAdapter<MacroToMicroGridFun<TGV,RF,Evl,3>> >(mu2GridFun,"mu2"));
    vtkwriter.addVertexData(std::make_shared<
      Dune::PDELab::VTKGridFunctionAdapter<MacroToMicroGridFun<TGV,RF,Evl,4>> >(vyGridFun,"vy"));
    vtkwriter.addVertexData(std::make_shared<
      Dune::PDELab::VTKGridFunctionAdapter<MacroToMicroGridFun<TGV,RF,Evl,5>> >(vxGridFun,"vx"));
  }
};

//_________________________________________________________________________________________________________________________

template<typename IniVX, typename IniVY, typename IniPhi, typename IniMu, typename IniPhi2, typename IniMu2>
class Ini_micro
{
public:
  IniVX iniVx;
  IniVY iniVy;
  IniPhi iniPhi;
  IniMu iniMu;
  IniPhi2 iniPhi2;
  IniMu2 iniMu2;
  typedef Dune::PDELab::CompositeGridFunction<IniVX> IniNS;
  typedef Dune::PDELab::CompositeGridFunction<IniPhi,IniMu,IniPhi2,IniMu2,IniVY> IniCH;
  IniNS iniNS;
  IniCH iniCH;

  template<typename GV, typename Params>
  Ini_micro(const GV& gv, const Params& params) :
      iniVx(gv, params),
      iniVy(gv, params),
      iniPhi(gv, params),
      iniMu(gv, params),
      iniPhi2(gv, params),
      iniMu2(gv, params),
      iniNS(iniVx),
      iniCH(iniPhi,iniMu,iniPhi2,iniMu2,iniVy)
  {
  }
};

//_________________________________________________________________________________________________________________________

template<typename BdryVX, typename BdryVY, typename BdryPhi, typename BdryMu, typename BdryPhi2, typename BdryMu2>
class Bdry_micro
{
public:
  BdryVX bdryVx;
  BdryVY bdryVy;
  BdryPhi bdryPhi;
  BdryMu bdryMu;
  BdryPhi2 bdryPhi2;
  BdryMu2 bdryMu2;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryVX> BdryNS;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryPhi, BdryMu, BdryPhi2, BdryMu2, BdryVY> BdryCH;
  BdryNS bdryNS;
  BdryCH bdryCH;

  Bdry_micro() :
      bdryVx(),
      bdryVy(),
      bdryPhi(),
      bdryMu(),
      bdryPhi2(),
      bdryMu2(),
      bdryNS(bdryVx),
      bdryCH(bdryPhi,bdryMu,bdryPhi2,bdryMu2,bdryVy)
  {
  }
};

/*template<typename BdryV1, typename BdryV2, typename BdryP,
          typename BdryPhi, typename BdryMu, typename BdryPhi2, typename BdryMu2, typename BdryU>
class Bdry_ffs_chns_r_compositeV_2d_Explicit_Domain
{
public:
  BdryV1 bdryV1;
  BdryV2 bdryV2;
  BdryP bdryP;
  BdryPhi bdryPhi;
  BdryMu bdryMu;
  BdryPhi2 bdryPhi2;
  BdryMu2 bdryMu2;
  BdryU bdryU;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryV1,BdryV2>    BdryV;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryV, BdryP> BdryNS;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryPhi, BdryMu, BdryPhi2, BdryMu2, BdryU> BdryCH;
  BdryV bdryV;
  BdryNS bdryNS;
  BdryCH bdryCH;

  template<typename DomainSize>
  Bdry_ffs_chns_r_compositeV_2d_Explicit_Domain(DomainSize ds) :
      bdryV1(ds),
      bdryV2(ds),
      bdryP(ds),
      bdryPhi(ds),
      bdryMu(ds),
      bdryPhi2(ds),
      bdryMu2(ds),
      bdryU(ds),
      bdryV(bdryV1,bdryV2),
      bdryNS(bdryV,bdryP),
      bdryCH(bdryPhi,bdryMu,bdryPhi2,bdryMu2,bdryU)
  {
  }
};

template<typename RF>
class DomainSize
{
public:
  const RF bottom;
  const RF top;
  const RF left;
  const RF right;
  const RF domainEps;

  DomainSize(RF b, RF t, RF l, RF r, RF eps):
  bottom(b), top(t), left(l), right(r), domainEps(eps)
  {
  }
};
*/
