#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include<iostream>
// dune includes
#include<dune/common/parallel/mpihelper.hh>
#include<dune/common/parametertreeparser.hh>
#include<dune/common/timer.hh>
//#if HAVE_UG
//#include<dune/grid/uggrid.hh>
//#endif
#if HAVE_DUNE_ALUGRID
#include<dune/alugrid/grid.hh>
#include<dune/alugrid/dgf.hh>
#include<dune/grid/io/file/dgfparser/dgfparser.hh>
#endif
#include <dune/grid/geometrygrid/grid.hh>
// pdelab includes
#include<dune/pdelab/finiteelementmap/pkfem.hh>
#include<dune/pdelab/finiteelementmap/qkfem.hh>
#include <dune/xt/grid/view/periodic.hh>
#include "driver_mixed.hh"

//===============================================================
// Main program with grid setup
//===============================================================



int main(int argc, char** argv)
{
  #if !HAVE_SUPERLU
    std::cout << "Error: These examples work only if SuperLU is available." << std::endl;
    exit(1);
  #endif

  #if HAVE_UG
  try{
    // Maybe initialize Mpi
    Dune::MPIHelper&
      helper = Dune::MPIHelper::instance(argc, argv);
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout << "Parallel code run on "
                << helper.size() << " process(es)" << std::endl;

    // open ini file
    Dune::ParameterTree ptree;
    Dune::ParameterTreeParser ptreeparser;
    ptreeparser.readINITree("chns.ini",ptree);
    ptreeparser.readOptions(argc,argv,ptree);

    // read ini file
//    const int dim = ptree.get<int>("grid.dim");
    const int refinement = ptree.get<int>("domain.level");


    if (true/*dim==2*/)
        {

/*          typedef Dune::ALUGrid<2,2,Dune::simplex,Dune::conforming> Grid;
          std::string filename = ptree.get("domain.filename",
                                           "turbtube2d.msh");
          Dune::GridFactory<Grid> factory;
          std::vector<int> boundary_index_map;
          std::vector<int> element_index_map;
          if (helper.rank()==0)
            Dune::GmshReader<Grid>::read(factory,filename,boundary_index_map,element_index_map,true,false);
          //std::shared_ptr<Grid> xgridp(factory.createGrid());
          std::shared_ptr<Grid> ygridp(factory.createGrid());
*/

          //typedef Dune::UGGrid<2> Grid;

          //typedef Dune::OneDGrid Grid;

          typedef Dune::YaspGrid<1> Grid;
          typedef Dune::YaspGrid<2> TensorGrid;
          typedef Grid::ctype DF;

          int A = ptree.get("domain.XResolution", 10);
          int B = ptree.get("domain.YResolution", 10);

          Dune::FieldVector<DF,1> L(1);
          Dune::FieldVector<DF,2> LL(1);
          std::shared_ptr<Grid> xgridp(new Grid(L,std::array<int,1>{A}));
          std::shared_ptr<Grid> ygridp(new Grid(L,std::array<int,1>{B},std::bitset<1>(false),0));
          std::shared_ptr<TensorGrid> tensorgridp(new TensorGrid(LL,std::array<int,2>{A,B}));

          typedef Grid::LeafGridView GV;
          typedef TensorGrid::LeafGridView TGV;
          GV xgv = xgridp->leafGridView();
          GV ygv = ygridp->leafGridView();
          TGV tgv = tensorgridp->leafGridView();

          std::cout << "OK" << std::endl;

          using ES = Dune::PDELab::NonOverlappingEntitySet<GV>;
          ES y_es(ygv);

          const int k = 2;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,1> V_FEM;
          const int l = 2;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,1> PHI_FEM;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,1> P_FEM;
          typedef Dune::PDELab::PkLocalFiniteElementMap<TGV,DF,double,1> Tensor_FEM;
          V_FEM vFem(y_es);
          P_FEM pFem(y_es);
          PHI_FEM phiFem(y_es);
          Tensor_FEM tensorFem(tgv);

          driver_mixed(xgv, pFem, A, ygv, y_es, tgv, tensorFem,vFem,phiFem, ptree, helper);
        }
      }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
    return 1;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
    return 1;
  }
  #endif
}
