#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "myincludes.hh"

// include all components making up this tutorial


#include "chns_driver.hh"

//===============================================================
// Main program with grid setup
//===============================================================




int main(int argc, char** argv)
{
  #if !HAVE_SUPERLU
    std::cout << "Error: These examples work only if SuperLU is available." << std::endl;
    exit(1);
  #endif

  #if HAVE_UG
  try{
    // Maybe initialize Mpi
    Dune::MPIHelper&
      helper = Dune::MPIHelper::instance(argc, argv);
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout << "Parallel code run on "
                << helper.size() << " process(es)" << std::endl;

    // open ini file
    Dune::ParameterTree ptree;
    Dune::ParameterTreeParser ptreeparser;
    ptreeparser.readINITree("2f0s.ini",ptree);
    ptreeparser.readOptions(argc,argv,ptree);

    // read ini file
//    const int dim = ptree.get<int>("grid.dim");
//    const int refinement = ptree.get<int>("domain.level");


    if (true/*dim==2*/)
        {

          typedef Dune::ALUGrid<2,2,Dune::simplex,Dune::conforming> Grid;
          //typedef Dune::UGGrid<2> Grid;
          typedef Grid::ctype DF;

          std::vector<int> boundary_index_map;
          std::vector<int> element_index_map;

          std::string filename = ptree.get("domain.filename",
                                           "turbtube2d.msh");
          Dune::GridFactory<Grid> factory;
          if (helper.rank()==0)
            Dune::GmshReader<Grid>::read(factory,filename,boundary_index_map,element_index_map,true,false);
          std::shared_ptr<Grid> gridp(factory.createGrid());
          //gridp->globalRefine(refinement);
          //gridp->loadBalance();
          //std::cout << " after load balance /" << helper.rank() << "/ " << gridp->size(0) << std::endl;

          //typedef Grid::LeafGridView LGV;

          //std::bitset<2> periodic_directions(std::string("01"));
          //typedef Dune::XT::Grid::PeriodicGridView<LGV> GV;
          //GV gv(gridp->leafGridView(),periodic_directions);

          typedef Grid::LeafGridView GV;
          GV gv = gridp->leafGridView();

          std::cout << "OK" << std::endl;

          using ES = Dune::PDELab::NonOverlappingEntitySet<GV>;
          ES es(gv);

          const int k = 2;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,k> V_FEM;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,k-1> P_FEM;
          const int l = 1;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,l> PHI_FEM;

          V_FEM vFem(es);
          P_FEM pFem(es);
          PHI_FEM phiFem(es);

          chns_driver(*gridp, es,vFem,pFem,phiFem, ptree, helper);
          }

        }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
    return 1;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
    return 1;
  }
  #endif
}
